from typing import Any

class Forcefield:
    def __init__(self, name: int): ...
    def get_parameters(
        self,
        group: str,
        key: str | list[str],
        keys_are_atom_classes: bool = False,
    ) -> dict[str, float]: ...
    @staticmethod
    def get_generator(ff: Forcefield, gen_type: Any) -> Any: ...
    @property
    def lj14scale(self) -> float: ...
    @property
    def coulomb14scale(self) -> float: ...
